@@ -1,16 +1,16 @@
 /*
- * Copyright 2005-2006 Gentoo Foundation
+ * Copyright 2005-2007 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/dumpelf.c,v 1.16 2006/01/05 03:12:07 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/dumpelf.c,v 1.24 2008/01/17 04:37:19 solar Exp $
  *
- * Copyright 2005-2006 Ned Ludd        - <solar@gentoo.org>
- * Copyright 2005-2006 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2005-2007 Ned Ludd        - <solar@gentoo.org>
+ * Copyright 2005-2007 Mike Frysinger  - <vapier@gentoo.org>
  */
 
-#include "paxinc.h"
+static const char *rcsid = "$Id: dumpelf.c,v 1.24 2008/01/17 04:37:19 solar Exp $";
+const char * const argv0 = "dumpelf";
 
-static const char *rcsid = "$Id: dumpelf.c,v 1.16 2006/01/05 03:12:07 vapier Exp $";
-#define argv0 "dumpelf"
+#include "paxinc.h"
 
 /* prototypes */
 static void dumpelf(const char *filename, long file_cnt);
@@ -26,13 +26,9 @@ static void dump_rela(elfobj *elf, void *rela);
 static void usage(int status);
 static void parseargs(int argc, char *argv[]);
 
-
-
 /* variables to control behavior */
 static char be_verbose = 0;
 
-
-
 /* dump all internal elf info */
 static void dumpelf(const char *filename, long file_cnt)
 {
@@ -43,6 +39,8 @@ static void dumpelf(const char *filename, long file_cnt)
 	if ((elf = readelf(filename)) == NULL)
 		return;
 
+	printf("#include <elf.h>\n");
+
 	printf(
 		"\n"
 		"/*\n"
@@ -140,7 +138,7 @@ static void dump_ehdr(elfobj *elf, void *ehdr_void)
 	       /* EI_BRAND, ehdr->e_ident[EI_BRAND] */ \
 	); \
 	printf("\t.e_type      = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_type), get_elfetype(elf)); \
-	printf("\t.e_machine   = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_machine), get_elfemtype(EGET(ehdr->e_machine))); \
+	printf("\t.e_machine   = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_machine), get_elfemtype(elf)); \
 	printf("\t.e_version   = %-10i ,\n", (int)EGET(ehdr->e_version)); \
 	printf("\t.e_entry     = 0x%-8lX ,\n", (unsigned long)EGET(ehdr->e_entry)); \
 	printf("\t.e_phoff     = %-10li , /* (bytes into file) */\n", (unsigned long)EGET(ehdr->e_phoff)); \
@@ -209,7 +207,7 @@ static void dump_shdr(elfobj *elf, void *shdr_void, long shdr_cnt, char *name)
 		} \
 		case SHT_STRTAB: { \
 			char bool; \
-			printf("\n\t/* section dump:\n"); \
+			printf("\n\t/%c section dump:\n", '*'); \
 			bool = 1; \
 			if (type == SHT_PROGBITS) --data; \
 			for (i = 0; i < size; ++i) { \
@@ -228,7 +226,7 @@ static void dump_shdr(elfobj *elf, void *shdr_void, long shdr_cnt, char *name)
 		} \
 		case SHT_DYNSYM: { \
 			Elf##B##_Sym *sym = (Elf##B##_Sym*)data; \
-			printf("\n\t/* section dump:\n"); \
+			printf("\n\t/%c section dump:\n", '*'); \
 			for (i = 0; i < EGET(shdr->sh_size) / EGET(shdr->sh_entsize); ++i) { \
 				printf("\t * Elf%i_Sym sym%li = {\n", B, (long)i); \
 				printf("\t * \t.st_name  = %i,\n", (unsigned int)EGET(sym->st_name)); \
@@ -246,11 +244,11 @@ static void dump_shdr(elfobj *elf, void *shdr_void, long shdr_cnt, char *name)
 		default: { \
 			if (be_verbose <= 1) \
 				break; \
-			printf("\n\t/* section dump:\n"); \
+			printf("\n\t/%c section dump:\n", '*'); \
 			for (i = 0; i < size; ++i) { \
 				++data; \
 				if (i % 20 == 0) printf("\t * "); \
-				printf("%.2X ", *data); \
+				printf("%.2X ", *data); /* this line can cause segfaults */ \
 				if (i % 20 == 19) printf("\n"); \
 			} \
 			if (i % 20) printf("\n"); \
@@ -264,8 +262,6 @@ static void dump_shdr(elfobj *elf, void *shdr_void, long shdr_cnt, char *name)
 	DUMP_SHDR(64)
 }
 
-
-
 /* usage / invocation handling functions */
 #define PARSE_FLAGS "vhV"
 #define a_argument required_argument
@@ -291,7 +287,7 @@ static void usage(int status)
 	printf("Options:\n");
 	for (i = 0; long_opts[i].name; ++i)
 		if (long_opts[i].has_arg == no_argument)
-			printf("  -%c, --%-13s* %s\n", long_opts[i].val, 
+			printf("  -%c, --%-13s* %s\n", long_opts[i].val,
 			       long_opts[i].name, opts_help[i]);
 		else
 			printf("  -%c, --%-6s <arg> * %s\n", long_opts[i].val,
@@ -329,20 +325,19 @@ static void parseargs(int argc, char *argv[])
 
 	if (optind == argc)
 		err("Nothing to dump !?");
+
 	{
 	long file_cnt = 0;
+
 	while (optind < argc)
 		dumpelf(argv[optind++], file_cnt++);
 	}
 }
 
-
-
 int main(int argc, char *argv[])
 {
 	if (argc < 2)
 		usage(EXIT_FAILURE);
-	printf("#include <elf.h>\n");
 	parseargs(argc, argv);
 	return EXIT_SUCCESS;
 }
