@@ -1,16 +1,14 @@
 /*
- * Copyright 2003-2006 Gentoo Foundation
+ * Copyright 2003-2007 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.c,v 1.52 2006/12/11 03:31:54 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.c,v 1.58 2008/01/17 04:37:19 solar Exp $
  *
- * Copyright 2005-2006 Ned Ludd        - <solar@gentoo.org>
- * Copyright 2005-2006 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2005-2007 Ned Ludd        - <solar@gentoo.org>
+ * Copyright 2005-2007 Mike Frysinger  - <vapier@gentoo.org>
  */
 
 #include "paxinc.h"
 
-#define argv0 "paxelf"
-
 /*
  * Setup a bunch of helper functions to translate
  * binary defines into readable strings.
@@ -108,6 +106,15 @@ const char *get_elfetype(elfobj *elf)
 	return find_pairtype(elf_etypes, get_etype(elf));
 }
 
+const char *get_endian(elfobj *elf)
+{
+	if (elf->data[EI_DATA] == ELFDATA2LSB)
+		return (char *) "LE";
+	if (elf->data[EI_DATA] == ELFDATA2MSB)
+		return (char *) "BE";
+	return (char *) "??";
+}
+
 void print_etypes(FILE *stream)
 {
 	int i, wrap = 0;
@@ -134,7 +141,6 @@ int etype_lookup(const char *str)
 	return atoi(str);
 }
 
-
 /* translate elf EM_ defines */
 static pairtype elf_emtypes[] = {
 	QUERY(EM_NONE),
@@ -247,7 +253,6 @@ const char *get_elfemtype(elfobj *elf)
 	return find_pairtype(elf_emtypes, get_emtype(elf));
 }
 
-
 /* translate elf PT_ defines */
 static pairtype elf_ptypes[] = {
 	QUERY(PT_NULL),
@@ -396,9 +401,7 @@ elfobj *readelf_buffer(const char *filename, char *buffer, size_t buffer_len)
 	if (buffer == NULL || buffer_len < EI_NIDENT)
 		return NULL;
 
-	elf = (elfobj*)malloc(sizeof(*elf));
-	if (elf == NULL)
-		return NULL;
+	elf = xmalloc(sizeof(*elf));
 	memset(elf, 0x00, sizeof(*elf));
 
 	elf->fd = -1;
@@ -598,7 +601,7 @@ unsigned long pax_pf2hf_flags(unsigned long paxflags)
 	char *pf_flags = pax_short_pf_flags(paxflags);
 	size_t x, len = strlen(pf_flags);
 	for (x = 0; x < len; x++) {
-		switch(pf_flags[x]) {
+		switch (pf_flags[x]) {
 			case 'p':
 				flags |= HF_PAX_PAGEEXEC;
 				break;
