@@ -8,12 +8,15 @@
 /*
  * normal compile.
  *  cc -o pspax pspax.c
- * or with libcap. 
+ * or with libcap.
  *  cc -o pspax pspax.c -DWANT_SYSCAP -lcap
- *
  */
 
+static const char *rcsid = "$Id: pspax.c,v 1.40 2008/01/17 04:37:19 solar Exp $";
+const char * const argv0 = "pspax";
+
 #include "paxinc.h"
+#include <grp.h>
 
 #ifdef WANT_SYSCAP
 # undef _POSIX_SOURCE
@@ -24,16 +27,13 @@
 #endif
 
 #define PROC_DIR "/proc"
-static const char *rcsid = "$Id: pspax.c,v 1.31 2006/08/12 16:47:52 solar Exp $";
-#define argv0 "pspax"
-
-
 
 /* variables to control behavior */
 static char show_all = 0;
 static char verbose = 0;
 static char show_banner = 1;
 static char show_phdr = 0;
+static char show_addr = 0;
 static char noexec = 1;
 static char writeexec = 1;
 
@@ -68,7 +68,7 @@ static int get_proc_maps(pid_t pid)
 	FILE *fp;
 
 	snprintf(str, sizeof(str), PROC_DIR "/%u/maps", pid);
-	
+
 	if ((fp = fopen(str, "r")) == NULL)
 		return -1;
 
@@ -79,8 +79,8 @@ static int get_proc_maps(pid_t pid)
 				continue;
 			/* 0x0-0x0 rwxp fffff000 00:00 0 */
 			/* 0x0-0x0 R+W+XP fffff000 00:00 0 */
-			++p; // ' '
-			++p; // r
+			++p; /* ' ' */
+			++p; /*  r  */
 			if (*p == '+')
 				++p;
 			/* FIXME: all of wx, w+, +x, ++ indicate w|x */
@@ -105,7 +105,7 @@ static int print_executable_mappings(pid_t pid)
 	FILE *fp;
 
 	snprintf(str, sizeof(str), PROC_DIR "/%u/maps", pid);
-	
+
 	if ((fp = fopen(str, "r")) == NULL)
 		return -1;
 
@@ -116,8 +116,8 @@ static int print_executable_mappings(pid_t pid)
 				continue;
 			/* 0x0-0x0 rwxp fffff000 00:00 0 */
 			/* 0x0-0x0 R+W+XP fffff000 00:00 0 */
-			++p; // ' '
-			++p; // r
+			++p; /* ' ' */
+			++p; /*  r  */
 			if (*p == '+')
 				++p;
 			/* FIXME: all of wx, w+, +x, ++ indicate w|x */
@@ -135,12 +135,12 @@ static int print_executable_mappings(pid_t pid)
 }
 
 #ifdef __BOUNDS_CHECKING_ON
-#define NOTE_TO_SELF warn( \
+# define NOTE_TO_SELF warn( \
 	"This is bullshit but getpwuid() is leaking memory and I wasted a few hrs 1 day tracking it down in pspax\n" \
 	"Later on I forgot I tracked it down before and saw pspax leaking memory so I tracked it down all over again (silly me)\n" \
 	"Hopefully the getpwuid()/nis/nss/pam or whatever wont suck later on in the future.")
 #else
- #define NOTE_TO_SELF
+# define NOTE_TO_SELF
 #endif
 
 static struct passwd *get_proc_passwd(pid_t pid)
@@ -200,6 +200,25 @@ static char *get_pid_attr(pid_t pid)
 	return buf;
 }
 
+static char *get_pid_addr(pid_t pid)
+{
+	FILE *fp;
+	char *p;
+	char str[32];
+	static char buf[BUFSIZ];
+
+	memset(buf, 0, sizeof(buf));
+
+	snprintf(str, sizeof(str), PROC_DIR "/%u/ipaddr", pid);
+	if ((fp = fopen(str, "r")) == NULL)
+		return NULL;
+	if (fgets(buf, sizeof(buf), fp) != NULL)
+		if ((p = strchr(buf, '\n')) != NULL)
+			*p = 0;
+	fclose(fp);
+	return buf;
+}
+
 static const char *get_proc_type(pid_t pid)
 {
 	char fname[32];
@@ -265,17 +284,16 @@ static const char *get_proc_phdr(pid_t pid)
 	return ret;
 }
 
-
 static void pspax(const char *find_name)
 {
 	register DIR *dir;
 	register struct dirent *de;
 	pid_t pid;
 	pid_t ppid = show_pid;
-	int have_attr, wx;
+	int have_attr, have_addr, wx;
 	struct passwd *pwd;
 	struct stat st;
-	const char *pax, *type, *name, *caps, *attr;
+	const char *pax, *type, *name, *caps, *attr, *addr;
 	WRAP_SYSCAP(ssize_t length; cap_t cap_d;);
 
 	WRAP_SYSCAP(cap_d = cap_init());
@@ -293,9 +311,15 @@ static void pspax(const char *find_name)
 	else
 		have_attr = 0;
 
+	if ((access("/proc/self/ipaddr", R_OK) != (-1)) && show_addr)
+		have_addr = 1;
+	else
+		have_addr = 0;
+
 	if (show_banner)
-		printf("%-8s %-6s %-6s %-4s %-10s %-16s %-4s %-4s %s\n",
-		       "USER", "PID", "PAX", "MAPS", "ETYPE", "NAME", "CAPS", "ATTR", show_phdr ? "STACK LOAD" : "");
+		printf("%-8s %-6s %-6s %-4s %-10s %-16s %-4s %-4s %s %s\n",
+		       "USER", "PID", "PAX", "MAPS", "ETYPE", "NAME", "CAPS", have_attr ? "ATTR" : "",
+			have_addr ? "IPADDR" : "", show_phdr ? "STACK LOAD" : "");
 
 	while ((de = readdir(dir))) {
 		errno = 0;
@@ -326,6 +350,7 @@ static void pspax(const char *find_name)
 			type = get_proc_type(pid);
 			name = get_proc_name(pid);
 			attr = (have_attr ? get_pid_attr(pid) : NULL);
+			addr = (have_addr ? get_pid_addr(pid) : NULL);
 
 			if (show_uid != (-1) && pwd)
 				if (pwd->pw_uid != show_uid)
@@ -340,7 +365,7 @@ static void pspax(const char *find_name)
 			WRAP_SYSCAP(caps = cap_to_text(cap_d, &length));
 
 			if (show_all || type) {
-				printf("%-8s %-6d %-6s %-4s %-10s %-16s %-4s %s %s\n",
+				printf("%-8s %-6d %-6s %-4s %-10s %-16s %-4s %s %s %s\n",
 				       pwd  ? pwd->pw_name : "--------",
 				       pid,
 				       pax  ? pax  : "---",
@@ -348,7 +373,9 @@ static void pspax(const char *find_name)
 				       type ? type : "-------",
 				       name ? name : "-----",
 				       caps ? caps : " = ",
-				       attr ? attr : "-", show_phdr ? get_proc_phdr(pid) : "");
+				       attr ? attr : "",
+				       addr ? addr : "",
+				       show_phdr ? get_proc_phdr(pid) : "");
 				if (verbose && wx)
 					print_executable_mappings(pid);
 			}
@@ -362,14 +389,13 @@ static void pspax(const char *find_name)
 	closedir(dir);
 }
 
-
-
 /* usage / invocation handling functions */
-#define PARSE_FLAGS "aep:u:g:nwvBhV"
+#define PARSE_FLAGS "aeip:u:g:nwvBhV"
 #define a_argument required_argument
 static struct option const long_opts[] = {
 	{"all",       no_argument, NULL, 'a'},
 	{"header",    no_argument, NULL, 'e'},
+	{"ipaddr",    no_argument, NULL, 'i'},
 	{"pid",        a_argument, NULL, 'p'},
 	{"user",       a_argument, NULL, 'u'},
 	{"group",      a_argument, NULL, 'g'},
@@ -384,9 +410,10 @@ static struct option const long_opts[] = {
 static const char *opts_help[] = {
 	"Show all processes",
 	"Print GNU_STACK/PT_LOAD markings",
+	"Print ipaddr info if supported",
 	"Process ID/pid #",
-	"Process UID #",
-	"Process GID #",
+	"Process user/uid #",
+	"Process group/gid #",
 	"Only display w^x processes",
 	"Only display w|x processes",
 	"Be verbose about executable mappings",
@@ -404,11 +431,11 @@ static void usage(int status)
 	       "Usage: %s [options]\n\n", argv0);
 	fputs("Options:\n", stdout);
 	for (i = 0; long_opts[i].name; ++i)
-		printf("  -%c, --%-12s* %s\n", long_opts[i].val, 
+		printf("  -%c, --%-12s* %s\n", long_opts[i].val,
 		       long_opts[i].name, opts_help[i]);
 #ifdef MANLYPAGE
 	for (i = 0; long_opts[i].name; ++i)
-		printf(".TP\n\\fB\\-%c, \\-\\-%s\\fR\n%s\n", long_opts[i].val, 
+		printf(".TP\n\\fB\\-%c, \\-\\-%s\\fR\n%s\n", long_opts[i].val,
 		       long_opts[i].name, opts_help[i]);
 #endif
 	exit(status);
@@ -418,6 +445,9 @@ static void usage(int status)
 static void parseargs(int argc, char *argv[])
 {
 	int flag;
+	struct passwd *pwd = NULL;
+	struct  group *gwd = NULL;
+
 	opterr = 0;
 	while ((flag=getopt_long(argc, argv, PARSE_FLAGS, long_opts, NULL)) != -1) {
 		switch (flag) {
@@ -433,13 +463,31 @@ static void parseargs(int argc, char *argv[])
 		case 'B': show_banner = 0; break;
 		case 'a': show_all = 1; break;
 		case 'e': show_phdr = 1; break;
+		case 'i': show_addr = 1; break;
 		case 'p': show_pid = atoi(optarg); break;
-		case 'u': show_uid = atoi(optarg); break;
-		case 'g': show_gid = atoi(optarg); break;
 		case 'n': noexec = 1; writeexec = 0; break;
 		case 'w': noexec = 0; writeexec = 1; break;
 		case 'v': verbose++; break;
-
+		case 'u':
+			show_uid = atoi(optarg);
+			if (show_uid == 0 && (strcmp(optarg, "0") != 0)) {
+				pwd = getpwnam(optarg);
+				if (pwd)
+					show_uid = pwd->pw_uid;
+				else
+					err("unknown uid");
+			}
+			break;
+		case 'g':
+			show_gid = atoi(optarg);
+			if (show_gid == 0 && (strcmp(optarg, "0") != 0)) {
+				gwd = getgrnam(optarg);
+				if (gwd)
+					show_gid = gwd->gr_gid;
+				else
+					err("unknown gid");
+			}
+			break;
 		case ':':
 		case '?':
 			warn("Unknown option or missing parameter");
@@ -452,8 +500,6 @@ static void parseargs(int argc, char *argv[])
 	}
 }
 
-
-
 int main(int argc, char *argv[])
 {
 	char *name = NULL;
