@@ -1,48 +1,33 @@
 /*
- * Copyright 2003-2006 Gentoo Foundation
+ * Copyright 2003-2007 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.158 2006/08/25 06:32:08 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.189 2008/01/17 04:37:19 solar Exp $
  *
- * Copyright 2003-2006 Ned Ludd        - <solar@gentoo.org>
- * Copyright 2004-2006 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2003-2007 Ned Ludd        - <solar@gentoo.org>
+ * Copyright 2004-2007 Mike Frysinger  - <vapier@gentoo.org>
  */
 
+static const char *rcsid = "$Id: scanelf.c,v 1.189 2008/01/17 04:37:19 solar Exp $";
+const char * const argv0 = "scanelf";
+
 #include "paxinc.h"
 
-static const char *rcsid = "$Id: scanelf.c,v 1.158 2006/08/25 06:32:08 vapier Exp $";
-#define argv0 "scanelf"
-
 #define IS_MODIFIER(c) (c == '%' || c == '#' || c == '+')
 
-#define do_state(option, flag)		\
-	if (islower(option)) {		\
-		flags &= ~PF_##flag;	\
-		flags |= PF_NO##flag;	\
-	} else {			\
-		flags &= ~PF_NO##flag;	\
-		flags |= PF_##flag;	\
-	}
-
-
 /* prototypes */
 static int file_matches_list(const char *filename, char **matchlist);
 static int scanelf_elfobj(elfobj *elf);
 static int scanelf_elf(const char *filename, int fd, size_t len);
 static int scanelf_archive(const char *filename, int fd, size_t len);
-static void scanelf_file(const char *filename);
-static void scanelf_dir(const char *path);
+static int scanelf_file(const char *filename, const struct stat *st_cache);
+static int scanelf_dir(const char *path);
 static void scanelf_ldpath(void);
 static void scanelf_envpath(void);
 static void usage(int status);
 static char **get_split_env(const char *envvar);
 static void parseenv(void);
-static void parseargs(int argc, char *argv[]);
-static char *xstrdup(const char *s);
-static void *xmalloc(size_t size);
-static void *xrealloc(void *ptr, size_t size);
-static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n);
-#define xstrcat(dst,src,curr_len) xstrncat(dst,src,curr_len,0)
-static inline void xchrcat(char **dst, const char append, size_t *curr_len);
+static int parseargs(int argc, char *argv[]);
+static int rematch(const char *regex, const char *match, int cflags);
 
 /* variables to control behavior */
 static char match_etypes[126] = "";
@@ -54,6 +39,7 @@ static char scan_archives = 0;
 static char dir_recurse = 0;
 static char dir_crossmount = 1;
 static char show_pax = 0;
+static char show_perms = 0;
 static char show_phdr = 0;
 static char show_textrel = 0;
 static char show_rpath = 0;
@@ -63,6 +49,7 @@ static char show_bind = 0;
 static char show_soname = 0;
 static char show_textrels = 0;
 static char show_banner = 1;
+static char show_endian = 0;
 static char be_quiet = 0;
 static char be_verbose = 0;
 static char be_wewy_wewy_quiet = 0;
@@ -73,7 +60,7 @@ static char *find_section = NULL;
 static char *out_format = NULL;
 static char *search_path = NULL;
 static char fix_elf = 0;
-static char gmatch = 0;
+static char g_match = 0;
 static char use_ldcache = 0;
 
 static char **qa_textrels = NULL;
@@ -81,11 +68,74 @@ static char **qa_execstack = NULL;
 static char **qa_wx_load = NULL;
 
 int match_bits = 0;
+unsigned int match_perms = 0;
 caddr_t ldcache = 0;
 size_t ldcache_size = 0;
 unsigned long setpax = 0UL;
 
+int has_objdump = 0;
 
+static char *getstr_perms(const char *fname);
+static char *getstr_perms(const char *fname)
+{
+	struct stat st;
+	static char buf[8];
+
+	if ((stat(fname, &st)) == (-1))
+		return (char *) "";
+
+	snprintf(buf, sizeof(buf), "%o", st.st_mode);
+
+	return (char *) buf + 2;
+}
+
+/* find the path to a file by name */
+static char *which(const char *fname)
+{
+	static char fullpath[BUFSIZ];
+	char *path, *p;
+
+	path = getenv("PATH");
+	if (!path)
+		return NULL;
+
+	path = xstrdup(path);
+	while ((p = strrchr(path, ':')) != NULL) {
+		snprintf(fullpath, sizeof(fullpath), "%s/%s", p + 1, fname);
+		*p = 0;
+		if (access(fullpath, R_OK) != (-1)) {
+			free(path);
+			return (char *) fullpath;
+		}
+	}
+	free(path);
+	return NULL;
+}
+
+/* 1 on failure. 0 otherwise */
+static int rematch(const char *regex, const char *match, int cflags)
+{
+	regex_t preg;
+	int ret;
+
+	if ((match == NULL) || (regex == NULL))
+		return EXIT_FAILURE;
+
+	if ((ret = regcomp(&preg, regex, cflags))) {
+		char err[256];
+
+		if (regerror(ret, &preg, err, sizeof(err)))
+			fprintf(stderr, "regcomp failed: %s", err);
+		else
+			fprintf(stderr, "regcomp failed");
+
+		return EXIT_FAILURE;
+	}
+	ret = regexec(&preg, match, 0, NULL, 0);
+	regfree(&preg);
+
+	return ret;
+}
 
 /* sub-funcs for scanelf_file() */
 static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **tab)
@@ -149,7 +199,6 @@ static char *scanelf_file_pax(elfobj *elf, char *found_pax)
 	SHOW_PAX(64)
 	}
 
-
 	if (fix_elf && setpax) {
 		/* set the chpax settings */
 		if (elf->elf_class == ELFCLASS32) {
@@ -217,7 +266,7 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 				offset = 4; \
 				check_flags = PF_X; \
 			} else if (EGET(phdr[i].p_type) == PT_LOAD) { \
-				if (ehdr->e_type == ET_DYN || ehdr->e_type == ET_EXEC) \
+				if (EGET(ehdr->e_type) == ET_DYN || EGET(ehdr->e_type) == ET_EXEC) \
 					if (multi_load++ > max_pt_load) \
 						warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
 				if (file_matches_list(elf->filename, qa_wx_load)) \
@@ -285,6 +334,10 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 		return ret;
 }
 
+/*
+ * See if this ELF contains a DT_TEXTREL tag in any of its
+ * PT_DYNAMIC sections.
+ */
 static const char *scanelf_file_textrel(elfobj *elf, char *found_textrel)
 {
 	static const char *ret = "TEXTREL";
@@ -302,7 +355,7 @@ static const char *scanelf_file_textrel(elfobj *elf, char *found_textrel)
 	Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
 	Elf ## B ## _Off offset; \
 	for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
-		if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+		if (EGET(phdr[i].p_type) != PT_DYNAMIC || EGET(phdr[i].p_filesz) == 0) continue; \
 		offset = EGET(phdr[i].p_offset); \
 		if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
 		dyn = DYN ## B (elf->data + offset); \
@@ -324,6 +377,12 @@ static const char *scanelf_file_textrel(elfobj *elf, char *found_textrel)
 	else
 		return "   -   ";
 }
+
+/*
+ * Scan the .text section to see if there are any relocations in it.
+ * Should rewrite this to check PT_LOAD sections that are marked
+ * Executable rather than the section named '.text'.
+ */
 static char *scanelf_file_textrels(elfobj *elf, char *found_textrels, char *found_textrel)
 {
 	unsigned long s, r, rmax;
@@ -400,7 +459,7 @@ static char *scanelf_file_textrels(elfobj *elf, char *found_textrels, char *foun
 			if (sym && sym->st_name) \
 				printf("%s", (char*)(elf->data + EGET(strtab->sh_offset) + EGET(sym->st_name))); \
 			else \
-				printf("(memory/fake?)"); \
+				printf("(memory/data?)"); \
 			printf(" [0x%lX]", (unsigned long)r_offset); \
 			/* now try to find the closest symbol that this rel is probably in */ \
 			sym = SYM ## B (elf->data + EGET(symtab->sh_offset)); \
@@ -414,11 +473,35 @@ static char *scanelf_file_textrels(elfobj *elf, char *found_textrels, char *foun
 				++sym; \
 			} \
 			printf(" in "); \
-			if (func && func->st_name) \
-				printf("%s", (char*)(elf->data + EGET(strtab->sh_offset) + EGET(func->st_name))); \
-			else \
-				printf("(NULL: fake?)"); \
+			if (func && func->st_name) { \
+				const char *func_name = elf->data + EGET(strtab->sh_offset) + EGET(func->st_name); \
+				if (r_offset > EGET(func->st_size)) \
+					printf("(optimized out: previous %s)", func_name); \
+				else \
+					printf("%s", func_name); \
+			} else \
+				printf("(optimized out)"); \
 			printf(" [0x%lX]\n", (unsigned long)offset_tmp); \
+			if (be_verbose && has_objdump) { \
+				Elf ## B ## _Addr end_addr = offset_tmp + EGET(func->st_size); \
+				char *sysbuf; \
+				size_t syslen; \
+				const char sysfmt[] = "objdump -r -R -d -w -l --start-address=0x%lX --stop-address=0x%lX %s | grep --color -i -C 3 '.*[[:space:]]%lX:[[:space:]]*R_.*'\n"; \
+				syslen = sizeof(sysfmt) + strlen(elf->filename) + 3 * sizeof(unsigned long) + 1; \
+				sysbuf = xmalloc(syslen); \
+				if (end_addr < r_offset) \
+					/* not uncommon when things are optimized out */ \
+					end_addr = r_offset + 0x100; \
+				snprintf(sysbuf, syslen, sysfmt, \
+					(unsigned long)offset_tmp, \
+					(unsigned long)end_addr, \
+					elf->filename, \
+					(unsigned long)r_offset); \
+				fflush(stdout); \
+				system(sysbuf); \
+				fflush(stdout); \
+				free(sysbuf); \
+			} \
 		} \
 	} }
 	SHOW_TEXTRELS(32)
@@ -447,7 +530,7 @@ static void rpath_security_checks(elfobj *elf, char *item, const char *dt_type)
 			if (fstat(elf->fd, &st) != -1)
 				if ((st.st_mode & S_ISUID) || (st.st_mode & S_ISGID))
 					warnf("Security problem with %s='%s' in %s with mode set of %o",
-					      dt_type, item, elf->filename, st.st_mode & 07777);
+					      dt_type, item, elf->filename, (unsigned int) st.st_mode & 07777);
 			break;
 		default:
 			warnf("Maybe? sec problem with %s='%s' in %s", dt_type, item, elf->filename);
@@ -477,7 +560,7 @@ static void scanelf_file_rpath(elfobj *elf, char *found_rpath, char **ret, size_
 		/* Scan all the program headers */ \
 		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
 			/* Just scan dynamic headers */ \
-			if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+			if (EGET(phdr[i].p_type) != PT_DYNAMIC || EGET(phdr[i].p_filesz) == 0) continue; \
 			offset = EGET(phdr[i].p_offset); \
 			if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
 			/* Just scan dynamic RPATH/RUNPATH headers */ \
@@ -737,7 +820,7 @@ static const char *scanelf_file_needed_lib(elfobj *elf, char *found_needed, char
 		Elf ## B ## _Shdr *strtbl = SHDR ## B (strtbl_void); \
 		Elf ## B ## _Off offset; \
 		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
-			if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+			if (EGET(phdr[i].p_type) != PT_DYNAMIC || EGET(phdr[i].p_filesz) == 0) continue; \
 			offset = EGET(phdr[i].p_offset); \
 			if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
 			dyn = DYN ## B (elf->data + offset); \
@@ -759,7 +842,7 @@ static const char *scanelf_file_needed_lib(elfobj *elf, char *found_needed, char
 						} \
 						*found_needed = 1; \
 					} else { \
-						if (!strncmp(find_lib, needed, strlen( !gmatch ? needed : find_lib))) { \
+						if (!strncmp(find_lib, needed, strlen( !g_match ? needed : find_lib))) { \
 							*found_lib = 1; \
 							return (be_wewy_wewy_quiet ? NULL : needed); \
 						} \
@@ -812,7 +895,7 @@ static char *scanelf_file_bind(elfobj *elf, char *found_bind)
 		Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
 		Elf ## B ## _Off offset; \
 		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
-			if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+			if (EGET(phdr[i].p_type) != PT_DYNAMIC || EGET(phdr[i].p_filesz) == 0) continue; \
 			dynamic = 1; \
 			offset = EGET(phdr[i].p_offset); \
 			if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
@@ -834,7 +917,8 @@ static char *scanelf_file_bind(elfobj *elf, char *found_bind)
 
 	if (be_wewy_wewy_quiet) return NULL;
 
-	if (be_quiet && !fstat(elf->fd, &s) && !(s.st_mode & S_ISUID || s.st_mode & S_ISGID)) {
+	/* don't output anything if quiet mode and the ELF is static or not setuid */
+	if (be_quiet && (!dynamic || (!fstat(elf->fd, &s) && !(s.st_mode & (S_ISUID|S_ISGID))))) {
 		return NULL;
 	} else {
 		*found_bind = 1;
@@ -860,10 +944,10 @@ static char *scanelf_file_soname(elfobj *elf, char *found_soname)
 		Elf ## B ## _Shdr *strtbl = SHDR ## B (strtbl_void); \
 		Elf ## B ## _Off offset; \
 		/* only look for soname in shared objects */ \
-		if (ehdr->e_type != ET_DYN) \
+		if (EGET(ehdr->e_type) != ET_DYN) \
 			return NULL; \
 		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
-			if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+			if (EGET(phdr[i].p_type) != PT_DYNAMIC || EGET(phdr[i].p_filesz) == 0) continue; \
 			offset = EGET(phdr[i].p_offset); \
 			if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
 			dyn = DYN ## B (elf->data + offset); \
@@ -887,6 +971,7 @@ static char *scanelf_file_soname(elfobj *elf, char *found_soname)
 
 	return NULL;
 }
+
 static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 {
 	unsigned long i;
@@ -918,13 +1003,13 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 					continue; \
 				} \
 				/* debug display ... show all symbols and some extra info */ \
-				if (*ret == '*') { \
-					printf("%s(%s) %5lX %15s %s\n", \
+				if (g_match ? rematch(ret, symname, REG_EXTENDED) == 0 : *ret == '*') { \
+					printf("%s(%s) %5lX %15s %s %s\n", \
 					       ((*found_sym == 0) ? "\n\t" : "\t"), \
 					       elf->base_filename, \
 					       (unsigned long)sym->st_size, \
 					       get_elfstttype(sym->st_info), \
-					       symname); \
+					       sym->st_shndx == SHN_UNDEF ? "U" : "D", symname); \
 					*found_sym = 1; \
 				} else { \
 					/* allow the user to specify a comma delimited list of symbols to search for */ \
@@ -982,7 +1067,6 @@ break_out:
 		return (char *)" - ";
 }
 
-
 static char *scanelf_file_sections(elfobj *elf, char *found_section)
 {
 	if (!find_section)
@@ -1017,8 +1101,8 @@ static char *scanelf_file_sections(elfobj *elf, char *found_section)
 static int scanelf_elfobj(elfobj *elf)
 {
 	unsigned long i;
-	char found_pax, found_phdr, found_relro, found_load, found_textrel, 
-	     found_rpath, found_needed, found_interp, found_bind, found_soname, 
+	char found_pax, found_phdr, found_relro, found_load, found_textrel,
+	     found_rpath, found_needed, found_interp, found_bind, found_soname,
 	     found_sym, found_lib, found_file, found_textrels, found_section;
 	static char *out_buffer = NULL;
 	static size_t out_len;
@@ -1037,7 +1121,7 @@ static int scanelf_elfobj(elfobj *elf)
 	/* init output buffer */
 	if (!out_buffer) {
 		out_len = sizeof(char) * 80;
-		out_buffer = (char*)xmalloc(out_len);
+		out_buffer = xmalloc(out_len);
 	}
 	*out_buffer = '\0';
 
@@ -1058,6 +1142,7 @@ static int scanelf_elfobj(elfobj *elf)
 			case 'e': prints("STK/REL/PTL "); break;
 			case 't': prints("TEXTREL "); break;
 			case 'r': prints("RPATH "); break;
+			case 'M': prints("CLASS "); break;
 			case 'n': prints("NEEDED "); break;
 			case 'i': prints("INTERP "); break;
 			case 'b': prints("BIND "); break;
@@ -1066,6 +1151,9 @@ static int scanelf_elfobj(elfobj *elf)
 			case 'N': prints("LIB "); break;
 			case 'T': prints("TEXTRELS "); break;
 			case 'k': prints("SECTION "); break;
+			case 'a': prints("ARCH "); break;
+			case 'O': prints("PERM "); break;
+			case 'D': prints("ENDIAN "); break;
 			default: warnf("'%c' has no title ?", out_format[i]);
 			}
 		}
@@ -1125,6 +1213,9 @@ static int scanelf_elfobj(elfobj *elf)
 		case 't': out = scanelf_file_textrel(elf, &found_textrel); break;
 		case 'T': out = scanelf_file_textrels(elf, &found_textrels, &found_textrel); break;
 		case 'r': scanelf_file_rpath(elf, &found_rpath, &out_buffer, &out_len); break;
+		case 'M': out = get_elfeitype(EI_CLASS, elf->data[EI_CLASS]); break;
+		case 'D': out = get_endian(elf); break;
+		case 'O': out = getstr_perms(elf->filename); break;
 		case 'n':
 		case 'N': out = scanelf_file_needed_lib(elf, &found_needed, &found_lib, (out_format[i]=='N'), &out_buffer, &out_len); break;
 		case 'i': out = scanelf_file_interp(elf, &found_interp); break;
@@ -1132,6 +1223,7 @@ static int scanelf_elfobj(elfobj *elf)
 		case 'S': out = scanelf_file_soname(elf, &found_soname); break;
 		case 's': out = scanelf_file_sym(elf, &found_sym); break;
 		case 'k': out = scanelf_file_sections(elf, &found_section); break;
+		case 'a': out = get_elfemtype(elf); break;
 		default: warnf("'%c' has no scan code?", out_format[i]);
 		}
 		if (out) {
@@ -1187,7 +1279,7 @@ static int scanelf_elf(const char *filename, int fd, size_t len)
 		strncpy(sbuf, match_etypes, sizeof(sbuf));
 		if (strchr(match_etypes, ',') != NULL) {
 			char *p;
-			while((p = strrchr(sbuf, ',')) != NULL) {
+			while ((p = strrchr(sbuf, ',')) != NULL) {
 				*p = 0;
 				if (etype_lookup(p+1) == get_etype(elf))
 					goto label_ret;
@@ -1230,62 +1322,64 @@ static int scanelf_archive(const char *filename, int fd, size_t len)
 	return 0;
 }
 /* scan a file which may be an elf or an archive or some other magical beast */
-static void scanelf_file(const char *filename)
+static int scanelf_file(const char *filename, const struct stat *st_cache)
 {
-	struct stat st;
+	const struct stat *st = st_cache;
+	struct stat symlink_st;
 	int fd;
 
-	/* make sure 'filename' exists */
-	if (lstat(filename, &st) == -1) {
-		if (be_verbose > 2) printf("%s: does not exist\n", filename);
-		return;
-	}
-
 	/* always handle regular files and handle symlinked files if no -y */
-	if (S_ISLNK(st.st_mode)) {
-		if (!scan_symlink) return;
-		stat(filename, &st);
+	if (S_ISLNK(st->st_mode)) {
+		if (!scan_symlink) return 1;
+		stat(filename, &symlink_st);
+		st = &symlink_st;
 	}
-	if (!S_ISREG(st.st_mode)) {
+
+	if (!S_ISREG(st->st_mode)) {
 		if (be_verbose > 2) printf("%s: skipping non-file\n", filename);
-		return;
+		return 1;
 	}
 
+	if (match_perms) {
+		if ((st->st_mode | match_perms) != st->st_mode)
+			return 1;
+	}
 	if ((fd=open(filename, (fix_elf ? O_RDWR : O_RDONLY))) == -1)
-		return;
+		return 1;
 
-	if (scanelf_elf(filename, fd, st.st_size) == 1 && scan_archives)
+	if (scanelf_elf(filename, fd, st->st_size) == 1 && scan_archives)
 		/* if it isn't an ELF, maybe it's an .a archive */
-		scanelf_archive(filename, fd, st.st_size);
+		scanelf_archive(filename, fd, st->st_size);
 
 	close(fd);
+	return 0;
 }
 
 /* scan a directory for ET_EXEC files and print when we find one */
-static void scanelf_dir(const char *path)
+static int scanelf_dir(const char *path)
 {
 	register DIR *dir;
 	register struct dirent *dentry;
 	struct stat st_top, st;
 	char buf[__PAX_UTILS_PATH_MAX];
 	size_t pathlen = 0, len = 0;
+	int ret = 0;
 
 	/* make sure path exists */
 	if (lstat(path, &st_top) == -1) {
 		if (be_verbose > 2) printf("%s: does not exist\n", path);
-		return;
+		return 1;
 	}
 
 	/* ok, if it isn't a directory, assume we can open it */
 	if (!S_ISDIR(st_top.st_mode)) {
-		scanelf_file(path);
-		return;
+		return scanelf_file(path, &st_top);
 	}
 
 	/* now scan the dir looking for fun stuff */
 	if ((dir = opendir(path)) == NULL) {
 		warnf("could not opendir %s: %s", path, strerror(errno));
-		return;
+		return 1;
 	}
 	if (be_verbose > 1) printf("%s: scanning dir\n", path);
 
@@ -1302,14 +1396,15 @@ static void scanelf_dir(const char *path)
 		snprintf(buf, sizeof(buf), "%s%s%s", path, (path[pathlen-1] == '/') ? "" : "/", dentry->d_name);
 		if (lstat(buf, &st) != -1) {
 			if (S_ISREG(st.st_mode))
-				scanelf_file(buf);
+				ret = scanelf_file(buf, &st);
 			else if (dir_recurse && S_ISDIR(st.st_mode)) {
 				if (dir_crossmount || (st_top.st_dev == st.st_dev))
-					scanelf_dir(buf);
+					ret = scanelf_dir(buf);
 			}
 		}
 	}
 	closedir(dir);
+	return ret;
 }
 
 static int scanelf_from_file(const char *filename)
@@ -1317,6 +1412,7 @@ static int scanelf_from_file(const char *filename)
 	FILE *fp = NULL;
 	char *p;
 	char path[__PAX_UTILS_PATH_MAX];
+	int ret = 0;
 
 	if (strcmp(filename, "-") == 0)
 		fp = stdin;
@@ -1327,11 +1423,11 @@ static int scanelf_from_file(const char *filename)
 		if ((p = strchr(path, '\n')) != NULL)
 			*p = 0;
 		search_path = path;
-		scanelf_dir(path);
+		ret = scanelf_dir(path);
 	}
 	if (fp != stdin)
 		fclose(fp);
-	return 0;
+	return ret;
 }
 
 #if defined(__GLIBC__) || defined(__UCLIBC__) || defined(__NetBSD__)
@@ -1354,7 +1450,7 @@ static int load_ld_cache_config(int i, const char *fname)
 		if ((p = strchr(path, '\n')) != NULL)
 			*p = 0;
 #ifdef __linux__
-		// recursive includes of the same file will make this segfault.
+		/* recursive includes of the same file will make this segfault. */
 		if ((memcmp(path, "include", 7) == 0) && isblank(path[7])) {
 			glob64_t gl;
 			size_t x;
@@ -1380,8 +1476,7 @@ static int load_ld_cache_config(int i, const char *fname)
 				}
 				globfree64 (&gl);
 				continue;
-			} else
-				abort();
+			}
 		}
 #endif
 		if (*path != '/')
@@ -1420,7 +1515,7 @@ static int load_ld_cache_config(int i, const char *fname)
 		return i;
 	}
 
-	b = (char*)malloc(hdr.dirlistlen+1);
+	b = xmalloc(hdr.dirlistlen + 1);
 	if (fread(b, 1, hdr.dirlistlen+1, fp) != hdr.dirlistlen+1) {
 		fclose(fp);
 		free(b);
@@ -1452,7 +1547,7 @@ static int load_ld_cache_config(int i, const char *fname)
 #endif
 
 /* scan /etc/ld.so.conf for paths */
-static void scanelf_ldpath()
+static void scanelf_ldpath(void)
 {
 	char scan_l, scan_ul, scan_ull;
 	int i = 0;
@@ -1476,7 +1571,7 @@ static void scanelf_ldpath()
 }
 
 /* scan env PATH for paths */
-static void scanelf_envpath()
+static void scanelf_envpath(void)
 {
 	char *path, *p;
 
@@ -1493,8 +1588,8 @@ static void scanelf_envpath()
 	free(path);
 }
 
-/* usage / invocation handling functions */
-#define PARSE_FLAGS "plRmyAXz:xetrnLibSs:k:gN:TaqvF:f:o:E:M:BhV"
+/* usage / invocation handling functions */ /* Free Flags: c d j u w C G H I J K P Q U W Y Z */
+#define PARSE_FLAGS "plRmyAXz:xetrnLibSs:k:gN:TaqvF:f:o:E:M:DO:BhV"
 #define a_argument required_argument
 static struct option const long_opts[] = {
 	{"path",      no_argument, NULL, 'p'},
@@ -1521,6 +1616,8 @@ static struct option const long_opts[] = {
 	{"textrels",  no_argument, NULL, 'T'},
 	{"etype",      a_argument, NULL, 'E'},
 	{"bits",       a_argument, NULL, 'M'},
+	{"endian",    no_argument, NULL, 'D'},
+	{"perms",      a_argument, NULL, 'O'},
 	{"all",       no_argument, NULL, 'a'},
 	{"quiet",     no_argument, NULL, 'q'},
 	{"verbose",   no_argument, NULL, 'v'},
@@ -1558,6 +1655,8 @@ static const char *opts_help[] = {
 	"Locate cause of TEXTREL",
 	"Print only ELF files matching etype ET_DYN,ET_EXEC ...",
 	"Print only ELF files matching numeric bits",
+	"Print Endianness",
+	"Print only ELF files matching octal permissions",
 	"Print all scanned info (-x -e -t -r -b)\n",
 	"Only output 'bad' things",
 	"Be verbose (can be specified more than once)",
@@ -1585,30 +1684,24 @@ static void usage(int status)
 			printf("  -%c, --%-7s <arg> * %s\n", long_opts[i].val,
 			       long_opts[i].name, opts_help[i]);
 
-	if (status != EXIT_SUCCESS)
-		exit(status);
-
-	puts("\nThe format modifiers for the -F option are:");
-	puts(" F Filename \tx PaX Flags \te STACK/RELRO");
-	puts(" t TEXTREL  \tr RPATH     \tn NEEDED");
-	puts(" i INTERP   \tb BIND      \ts symbol");
-	puts(" N library  \to Type      \tT TEXTRELs");
-	puts(" S SONAME   \tk section");
-	puts(" p filename (with search path removed)");
-	puts(" f filename (short name/basename)");
-	puts("Prefix each modifier with '%' (verbose) or '#' (silent)");
-
-	puts("\nELF Etypes:");
-	print_etypes(stdout);
-
+	puts("\nFor more information, see the scanelf(1) manpage");
 	exit(status);
 }
 
 /* parse command line arguments and preform needed actions */
-static void parseargs(int argc, char *argv[])
+#define do_pax_state(option, flag) \
+	if (islower(option)) { \
+		flags &= ~PF_##flag; \
+		flags |= PF_NO##flag; \
+	} else { \
+		flags &= ~PF_NO##flag; \
+		flags |= PF_##flag; \
+	}
+static int parseargs(int argc, char *argv[])
 {
 	int i;
 	const char *from_file = NULL;
+	int ret = 0;
 
 	opterr = 0;
 	while ((i=getopt_long(argc, argv, PARSE_FLAGS, long_opts, NULL)) != -1) {
@@ -1630,6 +1723,16 @@ static void parseargs(int argc, char *argv[])
 			break;
 		case 'M':
 			match_bits = atoi(optarg);
+			if (match_bits == 0) {
+				if (strcmp(optarg, "ELFCLASS32") == 0)
+					match_bits = 32;
+				if (strcmp(optarg, "ELFCLASS64") == 0)
+					match_bits = 64;
+			}
+			break;
+		case 'O':
+			if (sscanf(optarg, "%o", &match_perms) == (-1))
+				match_bits = 0;
 			break;
 		case 'o': {
 			if (freopen(optarg, "w", stdout) == NULL)
@@ -1643,7 +1746,7 @@ static void parseargs(int argc, char *argv[])
 		case 's': {
 			if (find_sym) warn("You prob don't want to specify -s twice");
 			find_sym = optarg;
-			versioned_symname = (char*)xmalloc(sizeof(char) * (strlen(find_sym)+1+1));
+			versioned_symname = xmalloc(sizeof(char) * (strlen(find_sym)+1+1));
 			sprintf(versioned_symname, "%s@", find_sym);
 			break;
 		}
@@ -1662,31 +1765,31 @@ static void parseargs(int argc, char *argv[])
 			unsigned long flags = (PF_NOEMUTRAMP | PF_NORANDEXEC);
 			size_t x;
 
-			for (x = 0 ; x < strlen(optarg); x++) {
-				switch(optarg[x]) {
+			for (x = 0; x < strlen(optarg); x++) {
+				switch (optarg[x]) {
 					case 'p':
 					case 'P':
-						do_state(optarg[x], PAGEEXEC);
+						do_pax_state(optarg[x], PAGEEXEC);
 						break;
 					case 's':
 					case 'S':
-						do_state(optarg[x], SEGMEXEC);
+						do_pax_state(optarg[x], SEGMEXEC);
 						break;
 					case 'm':
 					case 'M':
-						do_state(optarg[x], MPROTECT);
+						do_pax_state(optarg[x], MPROTECT);
 						break;
 					case 'e':
 					case 'E':
-						do_state(optarg[x], EMUTRAMP);
+						do_pax_state(optarg[x], EMUTRAMP);
 						break;
 					case 'r':
 					case 'R':
-						do_state(optarg[x], RANDMMAP);
+						do_pax_state(optarg[x], RANDMMAP);
 						break;
 					case 'x':
 					case 'X':
-						do_state(optarg[x], RANDEXEC);
+						do_pax_state(optarg[x], RANDEXEC);
 						break;
 					default:
 						break;
@@ -1701,7 +1804,7 @@ static void parseargs(int argc, char *argv[])
 					setpax = flags;
 			break;
 		}
-		case 'g': gmatch = 1; break;
+		case 'g': g_match = 1; break;
 		case 'L': use_ldcache = 1; break;
 		case 'y': scan_symlink = 0; break;
 		case 'A': scan_archives = 1; break;
@@ -1722,8 +1825,8 @@ static void parseargs(int argc, char *argv[])
 		case 'T': show_textrels = 1; break;
 		case 'q': be_quiet = 1; break;
 		case 'v': be_verbose = (be_verbose % 20) + 1; break;
-		case 'a': show_pax = show_phdr = show_textrel = show_rpath = show_bind = 1; break;
-
+		case 'a': show_perms = show_pax = show_phdr = show_textrel = show_rpath = show_bind = show_endian = 1; break;
+		case 'D': show_endian = 1; break;
 		case ':':
 			err("Option '%c' is missing parameter", optopt);
 		case '?':
@@ -1732,12 +1835,15 @@ static void parseargs(int argc, char *argv[])
 			err("Unhandled option '%c'; please report this", i);
 		}
 	}
-
+	if (show_textrels && be_verbose) {
+		if (which("objdump") != NULL)
+			has_objdump = 1;
+	}
 	/* let the format option override all other options */
 	if (out_format) {
 		show_pax = show_phdr = show_textrel = show_rpath = \
 		show_needed = show_interp = show_bind = show_soname = \
-		show_textrels = 0;
+		show_textrels = show_perms = show_endian = 0;
 		for (i = 0; out_format[i]; ++i) {
 			if (!IS_MODIFIER(out_format[i])) continue;
 
@@ -1752,6 +1858,10 @@ static void parseargs(int argc, char *argv[])
 			case 's': break;
 			case 'N': break;
 			case 'o': break;
+			case 'a': break;
+			case 'M': break;
+			case 'D': show_endian = 1; break;
+			case 'O': show_perms = 1; break;
 			case 'x': show_pax = 1; break;
 			case 'e': show_phdr = 1; break;
 			case 't': show_textrel = 1; break;
@@ -1762,7 +1872,7 @@ static void parseargs(int argc, char *argv[])
 			case 'S': show_soname = 1; break;
 			case 'T': show_textrels = 1; break;
 			default:
-				err("Invalid format specifier '%c' (byte %i)", 
+				err("Invalid format specifier '%c' (byte %i)",
 				    out_format[i], i+1);
 			}
 		}
@@ -1770,9 +1880,11 @@ static void parseargs(int argc, char *argv[])
 	/* construct our default format */
 	} else {
 		size_t fmt_len = 30;
-		out_format = (char*)xmalloc(sizeof(char) * fmt_len);
+		out_format = xmalloc(sizeof(char) * fmt_len);
 		if (!be_quiet)     xstrcat(&out_format, "%o ", &fmt_len);
 		if (show_pax)      xstrcat(&out_format, "%x ", &fmt_len);
+		if (show_perms)    xstrcat(&out_format, "%O ", &fmt_len);
+		if (show_endian)   xstrcat(&out_format, "%D ", &fmt_len);
 		if (show_phdr)     xstrcat(&out_format, "%e ", &fmt_len);
 		if (show_textrel)  xstrcat(&out_format, "%t ", &fmt_len);
 		if (show_rpath)    xstrcat(&out_format, "%r ", &fmt_len);
@@ -1803,7 +1915,7 @@ static void parseargs(int argc, char *argv[])
 		err("Nothing to scan !?");
 	while (optind < argc) {
 		search_path = argv[optind++];
-		scanelf_dir(search_path);
+		ret = scanelf_dir(search_path);
 	}
 
 	/* clean up */
@@ -1813,6 +1925,7 @@ static void parseargs(int argc, char *argv[])
 
 	if (ldcache != 0)
 		munmap(ldcache, ldcache_size);
+	return ret;
 }
 
 static char **get_split_env(const char *envvar)
@@ -1848,14 +1961,16 @@ static char **get_split_env(const char *envvar)
 	 * the envvals array of strings */
 	return envvals;
 }
-static void parseenv()
+
+static void parseenv(void)
 {
 	qa_textrels = get_split_env("QA_TEXTRELS");
 	qa_execstack = get_split_env("QA_EXECSTACK");
 	qa_wx_load = get_split_env("QA_WX_LOAD");
 }
+
 #ifdef __PAX_UTILS_CLEANUP
-static void cleanup()
+static void cleanup(void)
 {
 	free(out_format);
 	free(qa_textrels);
@@ -1864,14 +1979,13 @@ static void cleanup()
 }
 #endif
 
-
-
 int main(int argc, char *argv[])
 {
+	int ret;
 	if (argc < 2)
 		usage(EXIT_FAILURE);
 	parseenv();
-	parseargs(argc, argv);
+	ret = parseargs(argc, argv);
 	fclose(stdout);
 #ifdef __PAX_UTILS_CLEANUP
 	cleanup();
@@ -1879,54 +1993,8 @@ int main(int argc, char *argv[])
 	     "\t- 1 due to the out_buffer not being freed in scanelf_file()\n"
 	     "\t- 1 per QA_TEXTRELS/QA_EXECSTACK/QA_WX_LOAD");
 #endif
-	return EXIT_SUCCESS;
-}
-
-
-
-/* utility funcs */
-static char *xstrdup(const char *s)
-{
-	char *ret = strdup(s);
-	if (!ret) err("Could not strdup(): %s", strerror(errno));
 	return ret;
 }
-static void *xmalloc(size_t size)
-{
-	void *ret = malloc(size);
-	if (!ret) err("Could not malloc() %li bytes", (unsigned long)size);
-	return ret;
-}
-static void *xrealloc(void *ptr, size_t size)
-{
-	void *ret = realloc(ptr, size);
-	if (!ret) err("Could not realloc() %li bytes", (unsigned long)size);
-	return ret;
-}
-static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n)
-{
-	size_t new_len;
-
-	new_len = strlen(*dst) + strlen(src);
-	if (*curr_len <= new_len) {
-		*curr_len = new_len + (*curr_len / 2);
-		*dst = realloc(*dst, *curr_len);
-		if (!*dst)
-			err("could not realloc() %li bytes", (unsigned long)*curr_len);
-	}
-
-	if (n)
-		strncat(*dst, src, n);
-	else
-		strcat(*dst, src);
-}
-static inline void xchrcat(char **dst, const char append, size_t *curr_len)
-{
-	static char my_app[2];
-	my_app[0] = append;
-	my_app[1] = '\0';
-	xstrcat(dst, my_app, curr_len);
-}
 
 /* Match filename against entries in matchlist, return TRUE
  * if the file is listed */
